"""Notes widget."""
import os
from PyQt5 import QtWidgets, QtCore, QtGui

from cryspy import md_to_html

class WNotes(QtWidgets.QTextEdit):
    """WFunction class."""

    def __init__(self, parent=None):
        super(WNotes, self).__init__(parent)

        self.setAcceptRichText(True)
        self.setSizePolicy(
                QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                      QtWidgets.QSizePolicy.Expanding))
        self.setFont(QtGui.QFont("Courier", 8, QtGui.QFont.Normal))
        self.setAlignment(QtCore.Qt.AlignTop)
        self.setStyleSheet("background-color:white;")

        self.flag_text_changed = False
        self.textChanged.connect(lambda : setattr(self, "flag_text_changed",
                                                  True))
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)        
        self.customContextMenuRequested.connect(self.open_menu)
        self.file_name = None
        self.function_save_as = None
        self.function_open = None
        self.function_close_notes = None
        self.function_notes_to_html = None

    def set_function_save_as(self, function_save_as):
        """Set function for "Save notes as..." action."""
        self.function_save_as = function_save_as

    def set_function_open(self, function_open):
        """Set function for "Open notes" action."""
        self.function_open = function_open

    def set_function_close_notes(self, function_close_notes):
        """Set function for "Close notes" action."""
        self.function_close_notes = function_close_notes

    def set_function_notes_to_html(self, function_notes_to_html):
        """Internal function to display notes in Tab 'Notes'."""
        self.function_notes_to_html = function_notes_to_html

    def set_file_name(self, file_name:str):
        """Set object."""
        self.file_name = file_name

    def open_menu(self, position):
        """Context menu."""
        menu = QtWidgets.QMenu(self)
        open_notes = QtWidgets.QAction('Open notes', menu)
        open_notes.triggered.connect(self.open_new_notes)
        menu.addAction(open_notes)

        save_as = QtWidgets.QAction('Save notes as...', menu)
        save_as.triggered.connect(self.save_as)
        menu.addAction(save_as)

        close_notes = QtWidgets.QAction('Close notes', menu)
        close_notes.triggered.connect(self.close_notes)
        menu.addAction(close_notes)

        export_to_html = QtWidgets.QAction('Export to html', menu)
        export_to_html.triggered.connect(self.export_to_html)
        menu.addAction(export_to_html)

        menu.exec_(self.viewport().mapToGlobal(position))

    def close_notes(self):
        self.file_name = None
        self.clear()
        if self.function_close_notes is not None:
            self.function_close_notes()

    def open_new_notes(self):
        if self.function_open is not None:
            self.function_open()

    def open_notes(self):
        if self.file_name is None:
            return
        if os.path.isfile(self.file_name):
            with open(self.file_name, "r") as fid:
                s_text = fid.read()
            self.setText(s_text)
            if self.function_notes_to_html is not None:
                self.function_notes_to_html(md_to_html(s_text))

        self.flag_text_changed = False

    def save_as(self):
        if self.function_save_as is not None:
            self.function_save_as()
        self.flag_text_changed = False

    def save(self):

        s_text = self.toPlainText().strip()
        if s_text == "":
            return
        
        if self.function_notes_to_html is not None:
            self.function_notes_to_html(md_to_html(s_text))

        file_name = self.file_name 
        if file_name is None:
            return

        with open(file_name, "w") as fid:
            fid.write(s_text)
        self.flag_text_changed = False

    def focusOutEvent(self, event):
        """Submit changes just before focusing out."""
        QtWidgets.QTextEdit.focusOutEvent(self, event)
        if  self.flag_text_changed:
            self.save()
        
    def export_to_html(self):
        s_text = self.toPlainText().strip()
        if s_text == "":
            return
        file_name = self.file_name 
        if file_name is None:
            self.save_as()
            file_name = self.file_name 
        ind_point = file_name.rfind(".")
        file_name_pdf = f"{file_name[:ind_point]:}.html"
        s_html = md_to_html(s_text)
        with open(file_name_pdf, "w") as fid:
            fid.write(s_html)
        



"""WConsole class."""
from PyQt5 import QtWidgets, QtCore, QtGui


# class WEditCif(QtWidgets.QScrollArea):
class WEditCif(QtWidgets.QTextEdit):
    """WFunction class."""

    def __init__(self, parent=None):
        super(WEditCif, self).__init__(parent)

        self.setAcceptRichText(True)
        self.setSizePolicy(
                QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                      QtWidgets.QSizePolicy.Expanding))
        self.setFont(QtGui.QFont("Courier", 8, QtGui.QFont.Normal))
        self.setAlignment(QtCore.Qt.AlignTop)
        self.setStyleSheet("background-color:white;")
        self.text_changed = False
        self.textChanged.connect(lambda : setattr(self, "text_changed", True))
        self.object = None

    def set_object(self, obj):
        """Set object."""
        self.setText(obj.__repr__())
        self.setToolTip(obj.__doc__)
        self.object = obj
        self.text_changed = False

    def save_object(self):
        stext = self.toPlainText()
        obj2 = self.object.from_cif(stext)
        if obj2 is not None:
            self.object.copy_from(obj2)

    def focusOutEvent(self, event):
        """Submit changes just before focusing out."""
        QtWidgets.QTextEdit.focusOutEvent(self, event)
        if self.text_changed:
            self.save_object()
        


"""Functions to create docks for dock-area."""
from PyQt5 import QtWidgets

from cryspy import Pd, RhoChi, Pd2d, TOF

def w_for_interaction(obj, parent=None) -> tuple:
    """Give tuple of docks for Dockarea."""
    widget = None
    if isinstance(obj, RhoChi):
        widget = interaction_rhochi(obj, parent)
    elif isinstance(obj, (Pd, Pd2d, TOF)):
        widget = interaction_experiment(obj, parent)
    return widget

def interaction_rhochi(obj: RhoChi, parent=None):
    """Dock for RhoChi object."""
    widget = None
    experiments = obj.experiments()
    flag_none = True
    for experiment in experiments:
        widget = QtWidgets.QWidget(parent)
        layout = QtWidgets.QVBoxLayout()
        widg_exp = interaction_experiment(experiment, widget)
        if widg_exp != None:
            layout.addWidget(widg_exp)
            flag_none = False
    if flag_none:
        del(widget)
        return None
    return widget

def interaction_experiment(obj, parent=None):
    widget = None
    if obj.is_attribute("chi2"):
        chi2 = obj.chi2
        widget = QtWidgets.QGroupBox(obj.data_name, parent)
        grid_layout = QtWidgets.QGridLayout()

        cb_sum = QtWidgets.QCheckBox("sum", widget)
        cb_sum.setCheckState(2*int(chi2.sum))
        cb_sum.stateChanged.connect(
            lambda x: setattr(chi2, "sum", bool(x/2)))
        grid_layout.addWidget(cb_sum, 0, 0, 1, 1)

        cb_diff = QtWidgets.QCheckBox("diff", widget)
        cb_diff.setCheckState(2*int(chi2.diff))
        cb_diff.stateChanged.connect(
            lambda x: setattr(chi2, "diff", bool(x/2)))
        grid_layout.addWidget(cb_diff, 1, 0, 1, 1)

        cb_up = QtWidgets.QCheckBox("up", widget)
        cb_up.setCheckState(2*int(chi2.up))
        cb_up.stateChanged.connect(
            lambda x: setattr(chi2, "up", bool(x/2)))
        grid_layout.addWidget(cb_up, 0, 1, 1, 1)

        cb_down = QtWidgets.QCheckBox("down", widget)
        cb_down.setCheckState(2*int(chi2.down))
        cb_down.stateChanged.connect(
            lambda x: setattr(chi2, "down", bool(x/2)))
        grid_layout.addWidget(cb_down, 1, 1, 1, 1)
        widget.setLayout(grid_layout)
    return widget










"""WFunction class."""
from PyQt5 import QtWidgets, QtCore

try:
    from PyQt5.QtWebEngineWidgets import QWebEngineView
    flag_web_engine = True
except:
    flag_web_engine = False


from cryspy import LoopN, ItemN

from cryspy_editor.widgets.w_editcif import WEditCif
# from cryspy_editor.widgets.w_interactive import w_for_interaction

# from cryspy_editor.widgets.w_console import WConsole

from cryspy_editor.widgets.matplotlib import Graph
import matplotlib.pyplot as plt


class WBaseObj(QtWidgets.QTabWidget):
    """
    WBaseObj class.

    Attributes
    ----------
        - thread
    Mehtods
    -------
        - set_object
    """

    def __init__(self, parent=None):
        super(WBaseObj, self).__init__(parent)
        self.thread = None
        # self.w_notes = QWebEngineView(self)
        # self.w_notes.setHtml("<h1>Notes</h1>")
        # self.addTab(self.w_notes, "Notes")         

        # self.w_console = WConsole(self)
        # self.addTab(self.w_console, "Output")

    def set_thread(self, thread: QtCore.QThread):
        """Set thread."""
        self.thread = thread

    def setText(self, text: str):
        """Set text."""
        pass

    def write_to_object(self):
        """Write to object."""
        if self.object is None:
            return
        # stext = self.wlabel.toPlainText()
        # obj2 = self.object.from_cif(stext)
        # if obj2 is not None:
        #     self.object.copy_from(obj2)
        # run_function(pass_func, (), self.thread)

    def set_object(self, obj):
        """Set text."""
        if obj is None:
            return
        self.object = obj

        tab_text = ""
        if self.count() != 0:
            tab_text = str(self.tabText(self.currentIndex()))

        for ind_item in range(self.count()-1, -1, -1):
            self.removeTab(ind_item)
        plt.close()

        # Interactive tab
        # try:
        #     w_inter = w_for_interaction(obj, self)
        #     if w_inter is not None:
        #         self.insertTab(0, w_inter, "Interactive") 
        # except Exception as e:
        #     print("ERROR in interactive Tab")
        #     print(e)

        # RCIF tab
        if isinstance(obj, (LoopN, ItemN)):
            w_plain_text = WEditCif()
            w_plain_text.set_object(obj)
            self.addTab(w_plain_text, "RCIF format") 

        # Figure tab
        try:
            l_fig_ax = ([fig_ax for fig_ax in obj.plots() if fig_ax is not None])
        except Exception as e:
            l_fig_ax = []
            print("ERROR in obj.plots")
            print(e)

        for fig, ax in l_fig_ax:
            widget = QtWidgets.QWidget(self)
            layout = QtWidgets.QVBoxLayout()
            item_plot = Graph(fig, parent=widget)
            toolbar = item_plot.get_toolbar(parent=widget)
            layout.addWidget(toolbar)
            layout.addWidget(item_plot)
            widget.setLayout(layout)
            if isinstance(ax, tuple):
                s_text = f"Fig: {ax[0].title.get_text():}"
            else:
                s_text = f"Fig: {ax.title.get_text():}"
            if len(s_text) > 20:
                s_text = s_text[:20] + "..."
            self.addTab(widget, s_text)
            # self.insertTab(0, widget, s_text)

        # Report tab
        try:
            report_html = obj.report_html()
        except Exception as e:
            report_html = ""
            print("ERROR in obj.report_html")
            print(e)

        if report_html != "":
            if flag_web_engine:
                w_plain_text = QWebEngineView(self)
                w_plain_text.setHtml(report_html)
                self.addTab(w_plain_text, "View") 
                # self.insertTab(0, w_plain_text, "View")

        if self.count() == 0:
            q_label = QtWidgets.QLabel("No graphs and reports.")
            q_label.setSizePolicy(
                QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                      QtWidgets.QSizePolicy.Expanding))
            self.addTab(q_label, "")


        # if tab_text == "Notes":
        #     self.setCurrentIndex(0)
        # else:
        #     flag_first = True
        flag_first = True
        for ind_tab in range(self.count()):
            if tab_text == str(self.tabText(ind_tab)):
                self.setCurrentIndex(ind_tab)
                flag_first = False
                break
        if flag_first:
            self.setCurrentIndex(0)



"""WMethods class."""
from typing import Callable, Any, NoReturn
from PyQt5 import QtWidgets, QtGui, QtCore

from cryspy.A_functions_base.function_1_strings import value_error_to_string
from cryspy.A_functions_base.function_1_objects import variable_name_to_string

class WVariables(QtWidgets.QListWidget):
    """WFunction class."""

    def __init__(self, parent=None):
        super(WVariables, self).__init__(parent)
        # self.setWidgetResizable(True)

        self.object = None
        self.variable_names_refined = None
        self.variable_names_fixed = None

        # self.wlist = QtWidgets.QListWidget(self)
        # self.wlist.clicked.connect(self.open_menu)


        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)        
        self.customContextMenuRequested.connect(self.open_menu)
        self.itemDoubleClicked.connect(self.double_click)

        # self.setWidget(self.wlist)
        # self.wlist.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)        
        # self.wlist.customContextMenuRequested.connect(self.open_menu)
        # self.wlist.itemDoubleClicked.connect(self.double_click)

    def focusInEvent(self, event):
        self.get_variables()

    def set_func_object_clicked(self, func_object_clicked:
                                Callable[[Any], Any]) -> NoReturn:
        """Set function when object is clicked."""
        self.func_object_clicked = func_object_clicked

    def double_click(self, w_item):
        self.fix_refine_variable(w_item)

    def set_thread(self, thread):
        """Set text."""
        self.mythread = thread

    def set_object(self, obj):
        self.object = obj
        # self.wlist.clear()
        self.clear()
        self.variable_names_fixed = None
        self.get_variables()

    def get_variables(self):
        """Get methods."""
        obj = self.object
        if obj is None:
            return
        v_n_ref = obj.get_variable_names()

        self.variable_names_refined = v_n_ref
        v_n_fix = self.variable_names_fixed
        if v_n_fix is None:
            v_n_all = v_n_ref
        else:
            l_del = []
            for var_name in v_n_fix:
                try:
                    obj.get_variable_by_name(var_name)
                except AttributeError:
                    l_del.append(var_name)
            for del_name in l_del:
                v_n_fix.remove(del_name)

            for var_name in v_n_ref:
                if var_name in v_n_fix:
                    v_n_fix.remove(var_name)
            v_n_all = v_n_ref + v_n_fix

        qfont = QtGui.QFont()
        qfont.setBold(True)
        # self.wlist.clear()
        self.clear()
        for var_name in v_n_all:
            var_value = obj.get_variable_by_name(var_name)
            var_ref_name = var_name[:-1]+((f"{var_name[-1][0]:}_refinement",
                                             var_name[-1][1]), )
            var_sigma_name = var_name[:-1]+((f"{var_name[-1][0]:}_sigma",
                                             var_name[-1][1]), )
            var_sigma = obj.get_variable_by_name(var_sigma_name)
            var_ref = obj.get_variable_by_name(var_ref_name)
            
            name = variable_name_to_string(var_name, flag_html=False)
            name_sh = name.split(".")[-1]
            if var_ref:
                val_str = value_error_to_string(var_value, var_sigma)
                s_val = f"{name_sh:}: {val_str:}"
            else:
                s_val = f"{name_sh:}: {var_value:}"

            list_widget_item = QtWidgets.QListWidgetItem(s_val)
            list_widget_item.variable_name = var_name
            list_widget_item.setFont(qfont)
            # self.wlist.addItem(list_widget_item)
            self.addItem(list_widget_item)
        
        # self.wlist.setSortingEnabled(True)
        self.setSortingEnabled(True)

    def open_menu(self, position):
        """Context menu."""
        obj = self.object
        if obj is None:
            return
        menu = QtWidgets.QMenu(self)
        # w_list = self.wlist
        w_item = self.itemAt(position)
        if w_item is not None:
            # l_ind = find_tree_item_position(self, w_item)
            show_item = QtWidgets.QAction('Show item', menu)
            show_item.triggered.connect(
                lambda x: self.show_item(w_item))
            menu.addAction(show_item)


            fix_variable = QtWidgets.QAction('Delete variable', menu)
            fix_variable.triggered.connect(lambda x: self.delete_variable(
                w_item))
            menu.addAction(fix_variable)

            copy_variable = QtWidgets.QAction('Copy to clipboard', menu)
            copy_variable.triggered.connect(
                lambda x: self.copy_variable_to_clipboard(w_item))
            menu.addAction(copy_variable)
        else:
            fix_variables = QtWidgets.QAction('Fix all variables', menu)
            fix_variables.triggered.connect(lambda x: self.fix_variables())
            menu.addAction(fix_variables)

            copy_variables = QtWidgets.QAction('Copy to clipboard', menu)
            copy_variables.triggered.connect(
                lambda x: self.copy_variables_to_clipboard())
            menu.addAction(copy_variables)

        # menu.exec_(self.wlist.viewport().mapToGlobal(position))
        menu.exec_(self.viewport().mapToGlobal(position))

    def fix_variables(self):
        obj = self.object
        obj.fix_variables()
        # self.wlist.clear()
        self.clear()
        self.variable_names_fixed = None

    def copy_variables_to_clipboard(self):
        obj = self.object
        variable_names = obj.get_variable_names()
        l_name_str, l_val_str = [], []
        for var_name in variable_names:
            var_value = obj.get_variable_by_name(var_name)
            var_sigma_name = var_name[:-1]+((f"{var_name[-1][0]:}_sigma",
                                             var_name[-1][1]), )
            var_sigma = obj.get_variable_by_name(var_sigma_name)
            
            name = variable_name_to_string(var_name, flag_html=False)
            val_str = value_error_to_string(var_value, var_sigma)
            l_name_str.append(name)
            l_val_str.append(val_str)
        n_1 = max([len(h) for h in l_name_str])
        n_2 = max([len(h) for h in l_val_str])
        ls_out = [f"|{'parameter'.rjust(n_1):}|  {'value'.rjust(n_2):}|"]
        ls_out.append("|"+n_1*"-"+"|"+(n_2+2)*"-"+"|")
        ls_out.extend([f"|{h1.rjust(n_1):}|  {h2.rjust(n_2):}|" for h1, h2 in
                       zip(l_name_str, l_val_str)])

        cb = QtWidgets.QApplication.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText("\n".join(ls_out), mode=cb.Clipboard)

    def delete_variable(self, w_item):
        obj = self.object
        var_name = w_item.variable_name
        var_ref_name = var_name[:-1]+((f"{var_name[-1][0]:}_refinement",
                                       var_name[-1][1]), )
        try:
            obj.set_variable_by_name(var_ref_name, False)
            if self.variable_names_fixed is not None:
                if var_name in self.variable_names_fixed:
                    self.variable_names_fixed.remove(var_name)
        except AttributeError:
            if var_name in self.variable_names_fixed:
                self.variable_names_fixed.remove(var_name)
            if var_name in self.variable_names_refined:
                self.variable_names_refined.remove(var_name)
        # w_list = self.wlist
        # self.removeItemWidget(w_item)

    def show_item(self, w_item):
        obj = self.object
        var_name = w_item.variable_name
        item_name = var_name[:-1]
        item  = obj.get_variable_by_name(item_name)

        self.func_object_clicked(item)
        

    def fix_refine_variable(self, w_item):
        obj = self.object
        var_name = w_item.variable_name
        name = variable_name_to_string(var_name, flag_html=False)
        name_sh = name.split(".")[-1]
        var_value = obj.get_variable_by_name(var_name)
        var_ref_name = var_name[:-1]+((f"{var_name[-1][0]:}_refinement",
                                       var_name[-1][1]), )
        var_sigma_name = var_name[:-1]+((f"{var_name[-1][0]:}_sigma",
                                         var_name[-1][1]), )
        var_sigma = obj.get_variable_by_name(var_sigma_name)
        try:
            var_ref = obj.get_variable_by_name(var_ref_name)
            obj.set_variable_by_name(var_ref_name, not(var_ref))
            if var_ref:
                if self.variable_names_fixed is None:
                    self.variable_names_fixed = [var_name]
                elif not(var_name in self.variable_names_fixed):
                    self.variable_names_fixed.append(var_name)
                if var_name in self.variable_names_refined:
                    self.variable_names_refined.remove(var_name)
                w_item.setText(f"{name_sh:}: {var_value:}")
            else:
                if self.variable_names_fixed is not None:
                    if var_name in self.variable_names_fixed:
                        self.variable_names_fixed.remove(var_name)
                if not(var_name in self.variable_names_refined):
                    self.variable_names_refined.append(var_name)
                val_str = value_error_to_string(var_value, var_sigma)
                w_item.setText(f"{name_sh:}: {val_str:}")
                
        except AttributeError:
            if var_name in self.variable_names_fixed:
                self.variable_names_fixed.remove(var_name)
            if var_name in self.variable_names_refined:
                self.variable_names_refined.remove(var_name)
            # w_list = self.wlist
            self.removeItemWidget(w_item)

    def copy_variable_to_clipboard(self, w_item):
        cb = QtWidgets.QApplication.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText(w_item.text(), mode=cb.Clipboard)



"""Functions to create docks for dock-area."""
import numpy
from PyQt5 import QtWidgets, QtCore

from cryspy import AtomSiteL, Crystal, Cell, SpaceGroup, \
    PdInstrResolution, Diffrn, \
    PdBackgroundL, Pd, PdMeasL, RhoChi, Chi2, \
    Pd2dMeas, Pd2d, TOF, \
    Setup, PhaseL, Range, DiffrnRadiation, Pd2dBackground, \
    Pd2dInstrResolution, DiffrnOrientMatrix, DiffrnReflnL, Phase, \
    AtomElectronConfigurationL, MEM, AtomSiteScatL, \
    InversedHessian

from cryspy.A_functions_base.function_2_crystallography_base import \
    calc_atoms_in_unit_cell
from cryspy.A_functions_base.function_2_mem import \
    calc_index_atom_symmetry_closest_to_fract_xyz

from cryspy.B_parent_classes.cl_1_item import ItemN
from cryspy.B_parent_classes.cl_2_loop import LoopN
from cryspy.B_parent_classes.cl_3_data import DataN
from cryspy.B_parent_classes.cl_4_global import GlobalN

from cryspy.D_functions_item_loop.function_1_section_from_density_point \
    import calc_section_from_density_point

import matplotlib.pyplot as plt


def define_tool_buttons(obj, thread: QtCore.QThread) -> tuple:
    """Give actions."""
    w_actions = ()
    if isinstance(obj, RhoChi):
        w_actions = action_rhochi(obj, thread)
    elif isinstance(obj, MEM):
        w_actions = action_mem(obj, thread)
    elif isinstance(obj, Crystal):
        w_actions = action_crystal(obj, thread)
    elif isinstance(obj, Pd):
        w_actions = action_pd(obj, thread)
    elif isinstance(obj, Pd2d):
        w_actions = action_pd2d(obj, thread)
    elif isinstance(obj, TOF):
        w_actions = action_tof(obj, thread)
    elif isinstance(obj, Diffrn):
        w_actions = action_diffrn(obj, thread)
    elif isinstance(obj, AtomSiteScatL):
        w_actions = action_atom_site_scat_l(obj, thread)
    elif isinstance(obj, PdInstrResolution):
        w_actions = action_pd_instr_resolution(obj, thread)
    elif isinstance(obj, Cell):
        w_actions = action_cell(obj, thread)
    elif isinstance(obj, SpaceGroup):
        w_actions = action_space_group(obj, thread)
    elif isinstance(obj, DiffrnOrientMatrix):
        w_actions = action_diffrn_orient_matrix(obj, thread)
    elif isinstance(obj, DiffrnReflnL):
        w_actions = action_diffrn_refln_l(obj, thread)
    elif isinstance(obj, AtomSiteL):
        w_actions = action_atom_site_l(obj, thread)
    elif isinstance(obj, InversedHessian):
        w_actions = action_inversed_hessian(obj, thread)
    l_action = list(w_actions)
    l_action.extend(action_plots(obj))
    return tuple(l_action)


def action_atom_site_l(obj: AtomSiteL, thread: QtCore.QThread):
    """Give actions for AtomSiteL."""
    w_actions = []
    if obj.is_attribute("type_symbol"):
        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Show b_scat")
        qtb_1.clicked.connect(lambda: run_function(
            obj.report, (), thread))
        w_actions.append(qtb_1)
    return w_actions


def action_inversed_hessian(obj: InversedHessian, thread: QtCore.QThread):
    """Give actions for AtomSiteL."""
    w_actions = []
    if obj.is_defined():
        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Show correlation")
        qtb_1.clicked.connect(lambda: run_function(
            obj.report, (), thread))
        w_actions.append(qtb_1)
    return w_actions


def action_plots(obj):
    w_actions = []
    if not(isinstance(obj, (ItemN, LoopN, DataN, GlobalN))):
        return w_actions

    qtb_1 = QtWidgets.QToolButton()
    qtb_1.setText("Plots")

    def func_plot(obj):
        plots = obj.plots()
        for plot in plots:
            fig, ax = plot
            fig.show()
        return
    qtb_1.clicked.connect(lambda: func_plot(obj))
    w_actions.append(qtb_1)
    return w_actions


def action_atom_site_scat_l(obj: AtomSiteScatL, thread: QtCore.QThread):
    """Plot form factor."""
    w_actions = []
    if (obj.is_attribute("atom_type_scat") |
            obj.is_attribute("type_symbol")):
        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Plot form factors")
        sthovl = numpy.linspace(0, 2, 100)

        def func_plot(obj, sthovl, flag):
            fig, ax = obj.plot_form_factor()
            fig.show()
            return (fig, ax)
        qtb_1.clicked.connect(lambda: func_plot(obj, sthovl, False))
        w_actions.append(qtb_1)
    return w_actions


def action_pd_instr_resolution(obj: PdInstrResolution, thread: QtCore.QThread):
    """Plot resolution."""
    w_actions = []
    if (obj.is_defined()):
        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Plot resolution")
        ttheta = numpy.linspace(0, 140, 100)

        def func_plot(obj, ttheta):
            h_pv, eta, h_g, h_l, a_g, b_g, a_l, b_l = obj.calc_resolution(
                ttheta)
            plt.plot(ttheta, h_g, label="h_g")
            plt.plot(ttheta, h_l, label="h_l")
            plt.plot(ttheta, h_pv, label="h_pv")
            plt.legend(loc='upper right')
            plt.show()
            return
        qtb_1.clicked.connect(lambda: func_plot(obj, ttheta))
        w_actions.append(qtb_1)
    return w_actions


def action_cell(obj: Cell, thread: QtCore.QThread):
    """Plot resolution."""
    w_actions = []
    if (obj.is_defined()):
        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Cell report")
        qtb_1.clicked.connect(lambda: run_function(
            obj.report, (), thread))
        w_actions.append(qtb_1)
    return w_actions


def action_diffrn_orient_matrix(obj: DiffrnOrientMatrix,
                                thread: QtCore.QThread):
    """Plot resolution."""
    w_actions = []
    if (obj.is_defined()):
        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Show U matrix")

        def func_temp(obj):
            s_out = f"""Orientation matrix U:
          a* [c,  a*]        c
 X: {obj.u_11: 8.5f} {obj.u_12: 8.5f} {obj.u_13: 8.5f}
 Y: {obj.u_21: 8.5f} {obj.u_22: 8.5f} {obj.u_23: 8.5f}
 Z: {obj.u_31: 8.5f} {obj.u_32: 8.5f} {obj.u_33: 8.5f}

axis 'X' is along incident beam;
axis 'Z' is vertical direction."""
            return s_out

        qtb_1.clicked.connect(lambda: run_function(
            func_temp, (obj, ), thread))
        w_actions.append(qtb_1)
    return w_actions


def action_diffrn_refln_l(obj: DiffrnReflnL, thread: QtCore.QThread):
    """Give actions diffrn_refln."""
    w_actions = []
    if (obj.is_defined()):
        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Report experimental agreement factors")
        qtb_1.clicked.connect(lambda: run_function(
            obj.report_agreement_factor_exp, (), thread))
        w_actions.append(qtb_1)

        if (obj.is_attribute("fr") & obj.is_attribute("fr_sigma") &
                obj.is_attribute("fr_calc")):
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Report chi_sq exp.")
            qtb_1.clicked.connect(lambda: run_function(
                obj.report_chi_sq_exp, (), thread))
            w_actions.append(qtb_1)

            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Plot FR_exp/FR_mod")

            def func_plot(obj):
                plt.errorbar(obj.fr_calc, obj.fr, yerr=obj.fr_sigma, fmt=".")
                plt.plot(obj.fr, obj.fr)
                # plt.legend(loc='upper right')
                plt.show()
                return
            qtb_1.clicked.connect(lambda: func_plot(obj))

            w_actions.append(qtb_1)
    return w_actions


def action_space_group(obj: SpaceGroup, thread: QtCore.QThread):
    """Plot resolution."""
    w_actions = []
    if (obj.is_defined()):
        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Space Group report")
        qtb_1.clicked.connect(lambda: run_function(
            obj.report_space_group, (), thread))
        w_actions.append(qtb_1)
    return w_actions


def action_rhochi(obj: RhoChi, thread: QtCore.QThread):
    """Dock for RhoChi object."""
    w_actions = []
    crystals = obj.crystals()
    experiments = obj.experiments()

    flag_crystals = len(crystals) != 0
    flag_experiments = len(experiments) != 0
    flag_diffrn = any([isinstance(exp, Diffrn) for exp in experiments])

    # for experiment in experiments:
    #     if isinstance(experiment, Pd):
    #         w_actions_t = action_pd(experiment, thread)
    #         w_actions.extend(w_actions_t)
    #     elif isinstance(experiment, Pd2d):
    #         w_actions_t = action_pd2d(experiment, thread)
    #         w_actions.extend(w_actions_t)
    #     elif isinstance(experiment, Diffrn):
    #         w_actions_t = action_diffrn(experiment, thread)
    #         w_actions.extend(w_actions_t)

    # Action doc
    if (flag_crystals & flag_experiments & obj.is_defined()):
        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Calc. Chi square")
        qtb_1.clicked.connect(lambda: run_function(
            obj.calc_chi_sq, (True, ), thread))
        w_actions.append(qtb_1)

        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Refine")
        qtb_1.clicked.connect(lambda: run_function(
            obj.refine, (False, "BFGS",), thread))
        w_actions.append(qtb_1)

        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Estimate Inversed Hessian")
        qtb_1.clicked.connect(lambda: run_function(
            obj.estimate_inversed_hessian, (), thread))
        w_actions.append(qtb_1)

    elif not(flag_crystals & flag_experiments):
        if not flag_crystals:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add crystal")
            qtb_1.clicked.connect(lambda: add_items(obj, [
                Crystal(data_name="phase")], thread))
            w_actions.append(qtb_1)

        if not flag_experiments:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add diffrn")
            qtb_1.clicked.connect(lambda: add_items(obj, [
                Diffrn(data_name="mono")], thread))
            w_actions.append(qtb_1)

            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add pd")
            qtb_1.clicked.connect(lambda: add_items(obj, [
                Pd(data_name="powder1d")], thread))
            w_actions.append(qtb_1)

            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add pd2d")
            qtb_1.clicked.connect(lambda: add_items(obj, [
                Pd2d(data_name="powder2d")], thread))
            w_actions.append(qtb_1)
    else:
        qlabel = QtWidgets.QLabel(
            "To run calculations all items should be defined.")
        w_actions.append(qlabel)
        # layout_actions.addWidget(qlabel)
    if (flag_diffrn & flag_crystals):
        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Estimate F_M")
        qtb_1.clicked.connect(lambda: run_function(
            obj.estimate_f_mag_for_diffrn, (), thread))
        w_actions.append(qtb_1)
    if obj.is_attribute("inversed_hessian"):
        w_actions_t = action_inversed_hessian(obj.inversed_hessian, thread)
        w_actions.extend(w_actions_t)
    return w_actions


def action_mem(obj: MEM, thread: QtCore.QThread):
    """Dock for MEM object."""
    w_actions = []
    crystals = obj.crystals()
    experiments = obj.experiments()

    flag_crystals = len(crystals) != 0
    flag_experiments = len(experiments) != 0

    # for experiment in experiments:
    #     w_actions_temp = action_diffrn(experiment, thread)
    #     w_actions.extend(w_actions_temp)

    # Action doc
    if (flag_crystals & flag_experiments & obj.is_defined()):
        if flag_crystals:
            crystal = crystals[0]
            if not(crystal.is_attribute("atom_electron_configuration")):
                qtb_1 = QtWidgets.QToolButton()
                qtb_1.setText("Create AtomElectronConfiguration")
                qtb_1.clicked.connect(lambda: crystal.add_items([
                    AtomElectronConfigurationL()]))
                qtb_1.clicked.connect(lambda: run_function(pass_func, (),
                                                           thread))
                w_actions.append(qtb_1)

        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Create prior density")
        qtb_1.clicked.connect(lambda: run_function(
            obj.create_prior_density, (), thread))
        w_actions.append(qtb_1)

        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Calculate FR")
        qtb_1.clicked.connect(lambda: run_function(obj.calc_fr, (),
                                                   thread))
        w_actions.append(qtb_1)

        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Maximize entropy")
        qtb_1.clicked.connect(lambda: run_function(obj.maximize_entropy, (),
                                                   thread))
        w_actions.append(qtb_1)

        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Chi refinement")
        qtb_1.clicked.connect(lambda: run_function(obj.refine_susceptibility,
                                                   (), thread))
        w_actions.append(qtb_1)

        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Run cycle")
        qtb_1.clicked.connect(lambda: run_function(obj.make_cycle, (),
                                                   thread))
        w_actions.append(qtb_1)

        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Save to '.den' files")
        qtb_1.clicked.connect(lambda: run_function(obj.save_to_file_den, (),
                                                   thread))
        w_actions.append(qtb_1)

        if obj.is_attribute("section"):
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Plot sections")
    
            def func_plot(obj):
                crystal = obj.crystals()[0]
                space_group = crystal.space_group
                f_s_g_s = space_group.full_space_group_symop
                r_11 = numpy.array(f_s_g_s.r_11, dtype=float)
                r_12 = numpy.array(f_s_g_s.r_12, dtype=float)
                r_13 = numpy.array(f_s_g_s.r_13, dtype=float)
                r_21 = numpy.array(f_s_g_s.r_21, dtype=float)
                r_22 = numpy.array(f_s_g_s.r_22, dtype=float)
                r_23 = numpy.array(f_s_g_s.r_23, dtype=float)
                r_31 = numpy.array(f_s_g_s.r_31, dtype=float)
                r_32 = numpy.array(f_s_g_s.r_32, dtype=float)
                r_33 = numpy.array(f_s_g_s.r_33, dtype=float)
    
                r_ij = (r_11, r_12, r_13, r_21, r_22, r_23, r_31, r_32, r_33)
    
                b_1 = numpy.array(f_s_g_s.b_1, dtype=float)
                b_2 = numpy.array(f_s_g_s.b_2, dtype=float)
                b_3 = numpy.array(f_s_g_s.b_3, dtype=float)
    
                b_i = (b_1, b_2, b_3)
    
                atom_site = crystal.atom_site
                fract_x = atom_site.numpy_fract_x
                fract_y = atom_site.numpy_fract_y
                fract_z = atom_site.numpy_fract_z
                fract_xyz = (fract_x, fract_y, fract_z)
    
                atom_label = atom_site.numpy_label
    
                fract_uc_x, fract_uc_y, fract_uc_z, label_uc = \
                    calc_atoms_in_unit_cell(r_ij, b_i, fract_xyz, atom_label)
    
                cell = crystal.cell
                atom_site_susceptibility = crystal.atom_site_susceptibility
    
                section = obj.section[0]
                density_point = obj.density_point
                mem_parameters = obj.mem_parameters
    
                atom_x, atom_y, atom_label = section.calc_atoms(
                    cell, atom_site, f_s_g_s, distance_min=0.3)
    
                den_chi_section, den_b_section = \
                    calc_section_from_density_point(
                    section, density_point, mem_parameters, cell, f_s_g_s,
                    atom_site, atom_site_susceptibility)
    
                fract_atom_xyz = numpy.array(fract_xyz, dtype=float
                                             ).transpose()

                fract_sec_xyz = section.calc_fractions(cell, atom_site)
                fract_sec_xyz = numpy.transpose(numpy.array(fract_sec_xyz,
                                                            dtype=float))
    
                n_atom_index, n_symmetry, distance = \
                    calc_index_atom_symmetry_closest_to_fract_xyz(
                        fract_sec_xyz, fract_atom_xyz, r_ij, b_i, cell)
                n_at_2d = numpy.transpose(n_atom_index.reshape(
                    section.points_x, section.points_y))
    
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(4.2, 4.2),
                                               dpi=300)
                plt.set_cmap('Accent')
                ax1.imshow(n_at_2d, interpolation='bilinear',
                           extent=(-0.5*section.size_x, 0.5*section.size_x,
                                   -0.5*section.size_y, 0.5*section.size_y),
                           alpha=0.1, origin="lower")
    
                den_x = numpy.linspace(-0.5*section.size_x, 0.5*section.size_x,
                                       section.points_x)
                den_y = numpy.linspace(-0.5*section.size_y, 0.5*section.size_y,
                                       section.points_y)
    
                blk = '#000000'
                ax1.contour(den_x, den_y, den_chi_section.transpose(),
                            levels=[0.1, 0.5, 1., 5., 10., 50.],
                            colors=[blk, blk, blk, blk, blk, blk],
                            linewidths=0.5)
    
                ax1.plot(atom_x, atom_y, 'ko', ms=3)
                for _1, _2, _3 in zip(atom_x, atom_y, atom_label):
                    ax1.text(_1, _2, _3)
                ax1.set_title(
                    f"Tensor. Max is {den_chi_section.max():.1f}")
    
                # plt.set_cmap('RdBu')
                ax2.imshow(n_at_2d, interpolation='bilinear',
                           extent=(-0.5*section.size_x, 0.5*section.size_x,
                                   -0.5*section.size_y, 0.5*section.size_y),
                           alpha=0.1, origin="lower")
    
                hh = numpy.abs(den_b_section).max()
                rd = '#FF0000'
                ax2.contour(den_x, den_y, den_b_section.transpose(),
                            levels=[-50., -10., -5., -1., -0.5, -0.1, 
                                    0.1, 0.5, 1., 5., 10., 50.],
                            colors=[rd, rd, rd, rd, rd, rd,
                                    blk, blk, blk, blk, blk, blk],
                            linewidths=0.5)
                # ax2.imshow(den_b_section, interpolation='bilinear',
                #            extent=(-0.5*section.size_x, 0.5*section.size_x,
                #                    -0.5*section.size_y, 0.5*section.size_y),
                #            vmin=-hh, vmax=hh,
                #            alpha=1., origin="lower")
                ax2.set_title(f"2channel. Max is {hh:.1f}")
                ax2.plot(atom_x, atom_y, 'ko', ms=3)
                for _1, _2, _3 in zip(atom_x, atom_y, atom_label):
                    ax2.text(_1, _2, _3)
    
                plt.show()
    
                return
            qtb_1.clicked.connect(lambda: func_plot(obj))
            w_actions.append(qtb_1)

    elif not(flag_crystals & flag_experiments):
        if not flag_crystals:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add crystal")
            qtb_1.clicked.connect(lambda: add_items(obj, [
                Crystal(data_name="phase")], thread))
            w_actions.append(qtb_1)

        if not flag_experiments:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add diffrn")
            qtb_1.clicked.connect(lambda: add_items(obj, [
                Diffrn(data_name="mono")], thread))
            w_actions.append(qtb_1)

    else:
        qlabel = QtWidgets.QLabel(
            "To run calculations all items should be defined.")
        qlabel.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
                             QtWidgets.QSizePolicy.Expanding)
        w_actions.append(qlabel)

    return w_actions


def pass_func():
    """Pass function."""
    return


def add_items(obj, l_item, thread: QtCore.QThread):
    """Add items."""
    obj.add_items(l_item)
    thread.function = pass_func
    thread.arguments = ()
    thread.start()


def run_function(func, args, thread: QtCore.QThread):
    """Run function."""
    thread.function = func
    thread.arguments = args
    thread.start()


def action_pd(obj: Pd, thread: QtCore.QThread):
    """Form dock_pd.

    Based on
    --------
        - dock_proc
        - dock_meas
        - dock_chi2
        - dock_refine_ls
        - dock_peak
    """
    w_actions = []
    f_meas = obj.is_attribute("pd_meas")
    f_chi2 = obj.is_attribute("chi2")
    f_phase = obj.is_attribute("phase")

    l_pd_peak = []
    if f_phase:
        phase = obj.phase
        for item in phase.items:
            try:
                pd_peak = getattr(obj, f"pd_peak_{item.label.lower():}")
                l_pd_peak.append(pd_peak)
            except AttributeError:
                pass

    f_setup = obj.is_attribute("setup")
    f_pd_instr_resolution = obj.is_attribute("pd_instr_resolution")
    f_pd_background = obj.is_attribute("pd_background")
    f_range = obj.is_attribute("range")

    if not(f_chi2 & f_meas & f_setup & f_pd_instr_resolution & f_phase &
           f_pd_background & f_range):
        if not f_chi2:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add chi2")
            qtb_1.clicked.connect(lambda: add_items(obj, [Chi2()], thread))
            w_actions.append(qtb_1)

        if not f_meas:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add pd_meas")
            qtb_1.clicked.connect(lambda: add_items(obj, [PdMeasL()], thread))
            w_actions.append(qtb_1)

        if not f_setup:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add setup")
            qtb_1.clicked.connect(lambda: add_items(obj, [Setup()], thread))
            w_actions.append(qtb_1)

        if not f_pd_instr_resolution:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add pd_instr_resolution")
            qtb_1.clicked.connect(lambda: add_items(obj, [PdInstrResolution()],
                                                    thread))
            w_actions.append(qtb_1)

        if not f_phase:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add phase")
            vv = PhaseL()
            vv.items = [Phase(label="phase", igsize=0., scale=1.)]
            qtb_1.clicked.connect(lambda: add_items(obj, [vv], thread))
            w_actions.append(qtb_1)

        if not f_pd_background:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add pd_background")
            qtb_1.clicked.connect(lambda: add_items(obj, [PdBackgroundL()],
                                                    thread))
            w_actions.append(qtb_1)

        if not f_range:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add range")
            qtb_1.clicked.connect(lambda: add_items(obj, [Range(
                ttheta_min=2, ttheta_max=100.)], thread))
            w_actions.append(qtb_1)
    return w_actions


def action_tof(obj: TOF, thread: QtCore.QThread):
    """Form dock_pd.

    Based on
    --------
        - dock_proc
        - dock_meas
        - dock_chi2
        - dock_refine_ls
        - dock_peak
    """
    w_actions = []
    # f_meas = obj.is_attribute("pd_meas")
    # f_chi2 = obj.is_attribute("chi2")
    # f_phase = obj.is_attribute("phase")

    # l_pd_peak = []
    # if f_phase:
    #     phase = obj.phase
    #     for item in phase.items:
    #         try:
    #             pd_peak = getattr(obj, f"pd_peak_{item.label.lower():}")
    #             l_pd_peak.append(pd_peak)
    #         except AttributeError:
    #             pass

    # f_setup = obj.is_attribute("setup")
    # f_pd_instr_resolution = obj.is_attribute("pd_instr_resolution")
    # f_pd_background = obj.is_attribute("pd_background")
    # f_range = obj.is_attribute("range")

    # if not(f_chi2 & f_meas & f_setup & f_pd_instr_resolution & f_phase &
    #        f_pd_background & f_range):
    #     if not f_chi2:
    #         qtb_1 = QtWidgets.QToolButton()
    #         qtb_1.setText("Add chi2")
    #         qtb_1.clicked.connect(lambda: add_items(obj, [Chi2()], thread))
    #         w_actions.append(qtb_1)

    return w_actions


def action_crystal(obj: Crystal, thread: QtCore.QThread):
    """Docks for crystal."""
    w_actions = []
    if not(obj.is_attribute("cell")):
        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Add cell")
        qtb_1.clicked.connect(lambda: add_items(obj, [Cell()], thread))
        w_actions.append(qtb_1)
    else:
        cell = obj.cell
        w_actions.extend(action_cell(cell, thread))

    if not(obj.is_attribute("space_group")):
        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Add space group")
        qtb_1.clicked.connect(lambda: add_items(obj, [SpaceGroup(
            name_hm_alt="P 1")], thread))
        w_actions.append(qtb_1)
    else:
        space_group = obj.space_group
        w_actions.extend(action_space_group(space_group, thread))

    if not(obj.is_attribute("atom_site")):
        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Add atom site")
        qtb_1.clicked.connect(lambda: add_items(obj, [AtomSiteL()],
                                                thread))
        w_actions.append(qtb_1)
    else:
        atom_site = obj.atom_site
        w_actions.extend(action_atom_site_l(atom_site, thread))

    if obj.is_attribute("atom_site_susceptibility"):
        qtb_1 = QtWidgets.QToolButton()
        qtb_1.setText("Report main axes of magn.ellipsoids")
        qtb_1.clicked.connect(lambda: run_function(
            obj.report_main_axes_of_magnetization_ellipsoids, (), thread))
        w_actions.append(qtb_1)

    if obj.is_attribute("atom_site_scat"):
        atom_site_scat = obj.atom_site_scat
        w_actions.extend(action_atom_site_scat_l(atom_site_scat, thread))

    return w_actions


def action_pd2d(obj: Pd2d, thread: QtCore.QThread):
    """Actions for pd2d."""
    w_actions = []
    f_meas = obj.is_attribute("pd2d_meas")
    f_chi2 = obj.is_attribute("chi2")
    f_phase = obj.is_attribute("phase")

    f_setup = obj.is_attribute("setup")
    f_pd2d_instr_resolution = obj.is_attribute("pd2d_instr_resolution")
    f_phase = obj.is_attribute("phase")
    f_pd2d_background = obj.is_attribute("pd2d_background")
    f_range = obj.is_attribute("range")

    if not(f_chi2 & f_meas & f_setup & f_pd2d_instr_resolution & f_phase &
           f_pd2d_background & f_range):

        if not f_chi2:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add chi2")
            qtb_1.clicked.connect(lambda: add_items(obj, [Chi2()], thread))
            w_actions.append(qtb_1)

        if not f_meas:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add pd2d_meas")
            qtb_1.clicked.connect(lambda: add_items(obj, [Pd2dMeas()], thread))
            w_actions.append(qtb_1)

        if not f_setup:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add setup")
            qtb_1.clicked.connect(lambda: add_items(obj, [Setup()], thread))
            w_actions.append(qtb_1)

        if not f_pd2d_instr_resolution:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add pd2d_instr_resolution")
            qtb_1.clicked.connect(lambda: add_items(obj, [
                Pd2dInstrResolution()], thread))
            w_actions.append(qtb_1)

        if not f_phase:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add phase")
            vv = PhaseL()
            vv.items = [Phase(label="phase", igsize=0., scale=1.)]
            qtb_1.clicked.connect(lambda: add_items(obj, [vv], thread))
            w_actions.append(qtb_1)

        if not f_pd2d_background:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add pd2d_background")
            qtb_1.clicked.connect(lambda: add_items(obj, [Pd2dBackground()],
                                                    thread))
            w_actions.append(qtb_1)

        if not f_range:
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add range")
            qtb_1.clicked.connect(lambda: add_items(obj, [Range(
                ttheta_min=4., ttheta_max=80., phi_min=-10., phi_max=20.)],
                thread))
            w_actions.append(qtb_1)
    return w_actions


def action_diffrn(obj: Diffrn, thread: QtCore.QThread):
    """Actions for Diffrn objects."""
    w_actions = []

    f_setup = obj.is_attribute("setup")
    f_diffrn_radiation = obj.is_attribute("diffrn_radiation")
    f_diffrn_orient_matrix = obj.is_attribute("diffrn_orient_matrix")
    f_diffrn_refln = obj.is_attribute("diffrn_refln")
    f_phase = obj.is_attribute("phase")

    if not(f_setup & f_diffrn_radiation & f_diffrn_orient_matrix &
           f_diffrn_refln & f_phase):
        if not(f_setup):
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add setup")
            qtb_1.clicked.connect(lambda: add_items(obj, [Setup()], thread))
            w_actions.append(qtb_1)

        if not(f_diffrn_radiation):
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add diffrn_radiation")
            qtb_1.clicked.connect(lambda: add_items(
                obj, [DiffrnRadiation()], thread))
            w_actions.append(qtb_1)

        if not(f_diffrn_orient_matrix):
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add diffrn_orient_matrix")
            qtb_1.clicked.connect(lambda: add_items(obj, [DiffrnOrientMatrix(
                ub_11=1., ub_12=0., ub_13=0., ub_21=0., ub_22=1., ub_23=0.,
                ub_31=0., ub_32=0., ub_33=1.,)], thread))
            w_actions.append(qtb_1)

        if not(f_diffrn_refln):
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add diffrn_refln")
            qtb_1.clicked.connect(lambda: add_items(
                obj, [DiffrnReflnL()], thread))
            w_actions.append(qtb_1)

        if not(f_phase):
            qtb_1 = QtWidgets.QToolButton()
            qtb_1.setText("Add phase")
            qtb_1.clicked.connect(lambda: add_items(obj, [
                Phase(label="phase")], thread))
            w_actions.append(qtb_1)

    if f_diffrn_refln:
        diffrn_refln = obj.diffrn_refln
        w_actions.extend(action_diffrn_refln_l(diffrn_refln, thread))
    if f_diffrn_orient_matrix:
        diffrn_orient_matrix = obj.diffrn_orient_matrix
        w_actions.extend(action_diffrn_orient_matrix(
            diffrn_orient_matrix, thread))
    return w_actions



"""WMethods class."""
from typing import Callable, Any, NoReturn
import os
from PyQt5 import QtWidgets, QtGui, QtCore

from cryspy import load_packages, add_package, packages, delete_package

class WPackages(QtWidgets.QMainWindow):
    """WPackages class."""

    def __init__(self, parent=None):
        super(WPackages, self).__init__(parent)
        self.init_widget()
        self.setWindowTitle(
            'Additional packages of the CrysPy library')
        
    def init_widget(self):
        """Init widget."""
        cw = QtWidgets.QWidget(self)
        grid_layout = QtWidgets.QGridLayout()

        self.list_widget = QtWidgets.QListWidget(cw)
        l_package = [hh.strip() for hh in packages()]
        self.list_widget.addItems(l_package)
        self.pb_add = QtWidgets.QPushButton("Add package", cw)
        self.pb_add.clicked.connect(self.add_package)
        self.pb_del = QtWidgets.QPushButton("Delete secelcted package", cw)
        self.pb_del.clicked.connect(self.del_package)
        self.list_widget.itemDoubleClicked.connect(self.copy_path)
        # self.pb_load = QtWidgets.QPushButton("Load packages", cw)
        # self.pb_load.clicked.connect(self.load_package)

        grid_layout.addWidget(self.list_widget, 0, 0, 3, 1)
        grid_layout.addWidget(self.pb_add, 0, 1, 1, 1)
        grid_layout.addWidget(self.pb_del, 1, 1, 1, 1)
        # grid_layout.addWidget(self.pb_load, 2, 1, 1, 1)

        cw.setLayout(grid_layout)
        self.setCentralWidget(cw)

    def copy_path(self, w_item):
        f_dir = w_item.text()
        try:
            os.startfile(f_dir)
        except:
            cb = QtWidgets.QApplication.clipboard()
            cb.clear(mode=cb.Clipboard)
            cb.setText(f_dir, mode=cb.Clipboard)


    def add_package(self):
        f_dir = os.getcwd()
        f_name = QtWidgets.QFileDialog.getExistingDirectory(
            self, 'Directory of package', f_dir)
        if f_name=="":
            return None
        print("f_name: ", f_name)
        add_package(f_name)
        self.refresh()

    def del_package(self):
        s_package = self.list_widget.currentItem().text()
        delete_package(s_package)
        self.refresh()

    # def load_packages(self):
    #     load_packages()
    #     self.refresh()

    def refresh(self):
        load_packages()
        self.list_widget.clear()
        l_package = [hh.strip() for hh in packages()]
        self.list_widget.addItems(l_package)
